// Copyright 2017 Google LLC. All Rights Reserved.
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

// Code generated by protoc-gen-go. DO NOT EDIT.
// versions:
// 	protoc-gen-go v1.36.10
// 	protoc        v3.20.1
// source: client/configpb/multilog.proto

package configpb

import (
	protoreflect "google.golang.org/protobuf/reflect/protoreflect"
	protoimpl "google.golang.org/protobuf/runtime/protoimpl"
	timestamppb "google.golang.org/protobuf/types/known/timestamppb"
	reflect "reflect"
	sync "sync"
	unsafe "unsafe"
)

const (
	// Verify that this generated code is sufficiently up-to-date.
	_ = protoimpl.EnforceVersion(20 - protoimpl.MinVersion)
	// Verify that runtime/protoimpl is sufficiently up-to-date.
	_ = protoimpl.EnforceVersion(protoimpl.MaxVersion - 20)
)

// TemporalLogConfig is a set of LogShardConfig messages, whose
// time limits should be contiguous.
type TemporalLogConfig struct {
	state         protoimpl.MessageState `protogen:"open.v1"`
	Shard         []*LogShardConfig      `protobuf:"bytes,1,rep,name=shard,proto3" json:"shard,omitempty"`
	unknownFields protoimpl.UnknownFields
	sizeCache     protoimpl.SizeCache
}

func (x *TemporalLogConfig) Reset() {
	*x = TemporalLogConfig{}
	mi := &file_client_configpb_multilog_proto_msgTypes[0]
	ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
	ms.StoreMessageInfo(mi)
}

func (x *TemporalLogConfig) String() string {
	return protoimpl.X.MessageStringOf(x)
}

func (*TemporalLogConfig) ProtoMessage() {}

func (x *TemporalLogConfig) ProtoReflect() protoreflect.Message {
	mi := &file_client_configpb_multilog_proto_msgTypes[0]
	if x != nil {
		ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
		if ms.LoadMessageInfo() == nil {
			ms.StoreMessageInfo(mi)
		}
		return ms
	}
	return mi.MessageOf(x)
}

// Deprecated: Use TemporalLogConfig.ProtoReflect.Descriptor instead.
func (*TemporalLogConfig) Descriptor() ([]byte, []int) {
	return file_client_configpb_multilog_proto_rawDescGZIP(), []int{0}
}

func (x *TemporalLogConfig) GetShard() []*LogShardConfig {
	if x != nil {
		return x.Shard
	}
	return nil
}

// LogShardConfig describes the acceptable date range for a single shard of a temporal
// log.
type LogShardConfig struct {
	state protoimpl.MessageState `protogen:"open.v1"`
	Uri   string                 `protobuf:"bytes,1,opt,name=uri,proto3" json:"uri,omitempty"`
	// The log's public key in DER-encoded PKIX form.
	PublicKeyDer []byte `protobuf:"bytes,2,opt,name=public_key_der,json=publicKeyDer,proto3" json:"public_key_der,omitempty"`
	// not_after_start defines the start of the range of acceptable NotAfter
	// values, inclusive.
	// Leaving this unset implies no lower bound to the range.
	NotAfterStart *timestamppb.Timestamp `protobuf:"bytes,3,opt,name=not_after_start,json=notAfterStart,proto3" json:"not_after_start,omitempty"`
	// not_after_limit defines the end of the range of acceptable NotAfter values,
	// exclusive.
	// Leaving this unset implies no upper bound to the range.
	NotAfterLimit *timestamppb.Timestamp `protobuf:"bytes,4,opt,name=not_after_limit,json=notAfterLimit,proto3" json:"not_after_limit,omitempty"`
	unknownFields protoimpl.UnknownFields
	sizeCache     protoimpl.SizeCache
}

func (x *LogShardConfig) Reset() {
	*x = LogShardConfig{}
	mi := &file_client_configpb_multilog_proto_msgTypes[1]
	ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
	ms.StoreMessageInfo(mi)
}

func (x *LogShardConfig) String() string {
	return protoimpl.X.MessageStringOf(x)
}

func (*LogShardConfig) ProtoMessage() {}

func (x *LogShardConfig) ProtoReflect() protoreflect.Message {
	mi := &file_client_configpb_multilog_proto_msgTypes[1]
	if x != nil {
		ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
		if ms.LoadMessageInfo() == nil {
			ms.StoreMessageInfo(mi)
		}
		return ms
	}
	return mi.MessageOf(x)
}

// Deprecated: Use LogShardConfig.ProtoReflect.Descriptor instead.
func (*LogShardConfig) Descriptor() ([]byte, []int) {
	return file_client_configpb_multilog_proto_rawDescGZIP(), []int{1}
}

func (x *LogShardConfig) GetUri() string {
	if x != nil {
		return x.Uri
	}
	return ""
}

func (x *LogShardConfig) GetPublicKeyDer() []byte {
	if x != nil {
		return x.PublicKeyDer
	}
	return nil
}

func (x *LogShardConfig) GetNotAfterStart() *timestamppb.Timestamp {
	if x != nil {
		return x.NotAfterStart
	}
	return nil
}

func (x *LogShardConfig) GetNotAfterLimit() *timestamppb.Timestamp {
	if x != nil {
		return x.NotAfterLimit
	}
	return nil
}

var File_client_configpb_multilog_proto protoreflect.FileDescriptor

const file_client_configpb_multilog_proto_rawDesc = "" +
	"\n" +
	"\x1eclient/configpb/multilog.proto\x12\bconfigpb\x1a\x1fgoogle/protobuf/timestamp.proto\"C\n" +
	"\x11TemporalLogConfig\x12.\n" +
	"\x05shard\x18\x01 \x03(\v2\x18.configpb.LogShardConfigR\x05shard\"\xd0\x01\n" +
	"\x0eLogShardConfig\x12\x10\n" +
	"\x03uri\x18\x01 \x01(\tR\x03uri\x12$\n" +
	"\x0epublic_key_der\x18\x02 \x01(\fR\fpublicKeyDer\x12B\n" +
	"\x0fnot_after_start\x18\x03 \x01(\v2\x1a.google.protobuf.TimestampR\rnotAfterStart\x12B\n" +
	"\x0fnot_after_limit\x18\x04 \x01(\v2\x1a.google.protobuf.TimestampR\rnotAfterLimitBHZFgithub.com/google/certificate-transparency-go/client/multilog/configpbb\x06proto3"

var (
	file_client_configpb_multilog_proto_rawDescOnce sync.Once
	file_client_configpb_multilog_proto_rawDescData []byte
)

func file_client_configpb_multilog_proto_rawDescGZIP() []byte {
	file_client_configpb_multilog_proto_rawDescOnce.Do(func() {
		file_client_configpb_multilog_proto_rawDescData = protoimpl.X.CompressGZIP(unsafe.Slice(unsafe.StringData(file_client_configpb_multilog_proto_rawDesc), len(file_client_configpb_multilog_proto_rawDesc)))
	})
	return file_client_configpb_multilog_proto_rawDescData
}

var file_client_configpb_multilog_proto_msgTypes = make([]protoimpl.MessageInfo, 2)
var file_client_configpb_multilog_proto_goTypes = []any{
	(*TemporalLogConfig)(nil),     // 0: configpb.TemporalLogConfig
	(*LogShardConfig)(nil),        // 1: configpb.LogShardConfig
	(*timestamppb.Timestamp)(nil), // 2: google.protobuf.Timestamp
}
var file_client_configpb_multilog_proto_depIdxs = []int32{
	1, // 0: configpb.TemporalLogConfig.shard:type_name -> configpb.LogShardConfig
	2, // 1: configpb.LogShardConfig.not_after_start:type_name -> google.protobuf.Timestamp
	2, // 2: configpb.LogShardConfig.not_after_limit:type_name -> google.protobuf.Timestamp
	3, // [3:3] is the sub-list for method output_type
	3, // [3:3] is the sub-list for method input_type
	3, // [3:3] is the sub-list for extension type_name
	3, // [3:3] is the sub-list for extension extendee
	0, // [0:3] is the sub-list for field type_name
}

func init() { file_client_configpb_multilog_proto_init() }
func file_client_configpb_multilog_proto_init() {
	if File_client_configpb_multilog_proto != nil {
		return
	}
	type x struct{}
	out := protoimpl.TypeBuilder{
		File: protoimpl.DescBuilder{
			GoPackagePath: reflect.TypeOf(x{}).PkgPath(),
			RawDescriptor: unsafe.Slice(unsafe.StringData(file_client_configpb_multilog_proto_rawDesc), len(file_client_configpb_multilog_proto_rawDesc)),
			NumEnums:      0,
			NumMessages:   2,
			NumExtensions: 0,
			NumServices:   0,
		},
		GoTypes:           file_client_configpb_multilog_proto_goTypes,
		DependencyIndexes: file_client_configpb_multilog_proto_depIdxs,
		MessageInfos:      file_client_configpb_multilog_proto_msgTypes,
	}.Build()
	File_client_configpb_multilog_proto = out.File
	file_client_configpb_multilog_proto_goTypes = nil
	file_client_configpb_multilog_proto_depIdxs = nil
}
