// Code generated by protoc-gen-go. DO NOT EDIT.
// versions:
// 	protoc-gen-go v1.36.11
// 	protoc        v3.14.0
// source: github.com/moby/buildkit/solver/errdefs/errdefs.proto

package errdefs

import (
	pb "github.com/moby/buildkit/solver/pb"
	protoreflect "google.golang.org/protobuf/reflect/protoreflect"
	protoimpl "google.golang.org/protobuf/runtime/protoimpl"
	reflect "reflect"
	sync "sync"
	unsafe "unsafe"
)

const (
	// Verify that this generated code is sufficiently up-to-date.
	_ = protoimpl.EnforceVersion(20 - protoimpl.MinVersion)
	// Verify that runtime/protoimpl is sufficiently up-to-date.
	_ = protoimpl.EnforceVersion(protoimpl.MaxVersion - 20)
)

type Vertex struct {
	state         protoimpl.MessageState `protogen:"open.v1"`
	Digest        string                 `protobuf:"bytes,1,opt,name=digest,proto3" json:"digest,omitempty"`
	unknownFields protoimpl.UnknownFields
	sizeCache     protoimpl.SizeCache
}

func (x *Vertex) Reset() {
	*x = Vertex{}
	mi := &file_github_com_moby_buildkit_solver_errdefs_errdefs_proto_msgTypes[0]
	ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
	ms.StoreMessageInfo(mi)
}

func (x *Vertex) String() string {
	return protoimpl.X.MessageStringOf(x)
}

func (*Vertex) ProtoMessage() {}

func (x *Vertex) ProtoReflect() protoreflect.Message {
	mi := &file_github_com_moby_buildkit_solver_errdefs_errdefs_proto_msgTypes[0]
	if x != nil {
		ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
		if ms.LoadMessageInfo() == nil {
			ms.StoreMessageInfo(mi)
		}
		return ms
	}
	return mi.MessageOf(x)
}

// Deprecated: Use Vertex.ProtoReflect.Descriptor instead.
func (*Vertex) Descriptor() ([]byte, []int) {
	return file_github_com_moby_buildkit_solver_errdefs_errdefs_proto_rawDescGZIP(), []int{0}
}

func (x *Vertex) GetDigest() string {
	if x != nil {
		return x.Digest
	}
	return ""
}

type Source struct {
	state         protoimpl.MessageState `protogen:"open.v1"`
	Info          *pb.SourceInfo         `protobuf:"bytes,1,opt,name=info,proto3" json:"info,omitempty"`
	Ranges        []*pb.Range            `protobuf:"bytes,2,rep,name=ranges,proto3" json:"ranges,omitempty"`
	unknownFields protoimpl.UnknownFields
	sizeCache     protoimpl.SizeCache
}

func (x *Source) Reset() {
	*x = Source{}
	mi := &file_github_com_moby_buildkit_solver_errdefs_errdefs_proto_msgTypes[1]
	ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
	ms.StoreMessageInfo(mi)
}

func (x *Source) String() string {
	return protoimpl.X.MessageStringOf(x)
}

func (*Source) ProtoMessage() {}

func (x *Source) ProtoReflect() protoreflect.Message {
	mi := &file_github_com_moby_buildkit_solver_errdefs_errdefs_proto_msgTypes[1]
	if x != nil {
		ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
		if ms.LoadMessageInfo() == nil {
			ms.StoreMessageInfo(mi)
		}
		return ms
	}
	return mi.MessageOf(x)
}

// Deprecated: Use Source.ProtoReflect.Descriptor instead.
func (*Source) Descriptor() ([]byte, []int) {
	return file_github_com_moby_buildkit_solver_errdefs_errdefs_proto_rawDescGZIP(), []int{1}
}

func (x *Source) GetInfo() *pb.SourceInfo {
	if x != nil {
		return x.Info
	}
	return nil
}

func (x *Source) GetRanges() []*pb.Range {
	if x != nil {
		return x.Ranges
	}
	return nil
}

type Frontend struct {
	state         protoimpl.MessageState `protogen:"open.v1"`
	Name          string                 `protobuf:"bytes,1,opt,name=name,proto3" json:"name,omitempty"`     // frontend name e.g. dockerfile.v0 or gateway.v0
	Source        string                 `protobuf:"bytes,2,opt,name=source,proto3" json:"source,omitempty"` // used by the gateway frontend to identify the source, which corresponds to the image name
	unknownFields protoimpl.UnknownFields
	sizeCache     protoimpl.SizeCache
}

func (x *Frontend) Reset() {
	*x = Frontend{}
	mi := &file_github_com_moby_buildkit_solver_errdefs_errdefs_proto_msgTypes[2]
	ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
	ms.StoreMessageInfo(mi)
}

func (x *Frontend) String() string {
	return protoimpl.X.MessageStringOf(x)
}

func (*Frontend) ProtoMessage() {}

func (x *Frontend) ProtoReflect() protoreflect.Message {
	mi := &file_github_com_moby_buildkit_solver_errdefs_errdefs_proto_msgTypes[2]
	if x != nil {
		ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
		if ms.LoadMessageInfo() == nil {
			ms.StoreMessageInfo(mi)
		}
		return ms
	}
	return mi.MessageOf(x)
}

// Deprecated: Use Frontend.ProtoReflect.Descriptor instead.
func (*Frontend) Descriptor() ([]byte, []int) {
	return file_github_com_moby_buildkit_solver_errdefs_errdefs_proto_rawDescGZIP(), []int{2}
}

func (x *Frontend) GetName() string {
	if x != nil {
		return x.Name
	}
	return ""
}

func (x *Frontend) GetSource() string {
	if x != nil {
		return x.Source
	}
	return ""
}

type FrontendCap struct {
	state         protoimpl.MessageState `protogen:"open.v1"`
	Name          string                 `protobuf:"bytes,1,opt,name=name,proto3" json:"name,omitempty"`
	unknownFields protoimpl.UnknownFields
	sizeCache     protoimpl.SizeCache
}

func (x *FrontendCap) Reset() {
	*x = FrontendCap{}
	mi := &file_github_com_moby_buildkit_solver_errdefs_errdefs_proto_msgTypes[3]
	ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
	ms.StoreMessageInfo(mi)
}

func (x *FrontendCap) String() string {
	return protoimpl.X.MessageStringOf(x)
}

func (*FrontendCap) ProtoMessage() {}

func (x *FrontendCap) ProtoReflect() protoreflect.Message {
	mi := &file_github_com_moby_buildkit_solver_errdefs_errdefs_proto_msgTypes[3]
	if x != nil {
		ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
		if ms.LoadMessageInfo() == nil {
			ms.StoreMessageInfo(mi)
		}
		return ms
	}
	return mi.MessageOf(x)
}

// Deprecated: Use FrontendCap.ProtoReflect.Descriptor instead.
func (*FrontendCap) Descriptor() ([]byte, []int) {
	return file_github_com_moby_buildkit_solver_errdefs_errdefs_proto_rawDescGZIP(), []int{3}
}

func (x *FrontendCap) GetName() string {
	if x != nil {
		return x.Name
	}
	return ""
}

type CompatibilityFeature struct {
	state         protoimpl.MessageState `protogen:"open.v1"`
	Version       int64                  `protobuf:"varint,1,opt,name=version,proto3" json:"version,omitempty"`
	Feature       string                 `protobuf:"bytes,2,opt,name=feature,proto3" json:"feature,omitempty"`
	unknownFields protoimpl.UnknownFields
	sizeCache     protoimpl.SizeCache
}

func (x *CompatibilityFeature) Reset() {
	*x = CompatibilityFeature{}
	mi := &file_github_com_moby_buildkit_solver_errdefs_errdefs_proto_msgTypes[4]
	ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
	ms.StoreMessageInfo(mi)
}

func (x *CompatibilityFeature) String() string {
	return protoimpl.X.MessageStringOf(x)
}

func (*CompatibilityFeature) ProtoMessage() {}

func (x *CompatibilityFeature) ProtoReflect() protoreflect.Message {
	mi := &file_github_com_moby_buildkit_solver_errdefs_errdefs_proto_msgTypes[4]
	if x != nil {
		ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
		if ms.LoadMessageInfo() == nil {
			ms.StoreMessageInfo(mi)
		}
		return ms
	}
	return mi.MessageOf(x)
}

// Deprecated: Use CompatibilityFeature.ProtoReflect.Descriptor instead.
func (*CompatibilityFeature) Descriptor() ([]byte, []int) {
	return file_github_com_moby_buildkit_solver_errdefs_errdefs_proto_rawDescGZIP(), []int{4}
}

func (x *CompatibilityFeature) GetVersion() int64 {
	if x != nil {
		return x.Version
	}
	return 0
}

func (x *CompatibilityFeature) GetFeature() string {
	if x != nil {
		return x.Feature
	}
	return ""
}

type Subrequest struct {
	state         protoimpl.MessageState `protogen:"open.v1"`
	Name          string                 `protobuf:"bytes,1,opt,name=name,proto3" json:"name,omitempty"`
	unknownFields protoimpl.UnknownFields
	sizeCache     protoimpl.SizeCache
}

func (x *Subrequest) Reset() {
	*x = Subrequest{}
	mi := &file_github_com_moby_buildkit_solver_errdefs_errdefs_proto_msgTypes[5]
	ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
	ms.StoreMessageInfo(mi)
}

func (x *Subrequest) String() string {
	return protoimpl.X.MessageStringOf(x)
}

func (*Subrequest) ProtoMessage() {}

func (x *Subrequest) ProtoReflect() protoreflect.Message {
	mi := &file_github_com_moby_buildkit_solver_errdefs_errdefs_proto_msgTypes[5]
	if x != nil {
		ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
		if ms.LoadMessageInfo() == nil {
			ms.StoreMessageInfo(mi)
		}
		return ms
	}
	return mi.MessageOf(x)
}

// Deprecated: Use Subrequest.ProtoReflect.Descriptor instead.
func (*Subrequest) Descriptor() ([]byte, []int) {
	return file_github_com_moby_buildkit_solver_errdefs_errdefs_proto_rawDescGZIP(), []int{5}
}

func (x *Subrequest) GetName() string {
	if x != nil {
		return x.Name
	}
	return ""
}

type Solve struct {
	state    protoimpl.MessageState `protogen:"open.v1"`
	InputIDs []string               `protobuf:"bytes,1,rep,name=inputIDs,proto3" json:"inputIDs,omitempty"`
	MountIDs []string               `protobuf:"bytes,2,rep,name=mountIDs,proto3" json:"mountIDs,omitempty"`
	Op       *pb.Op                 `protobuf:"bytes,3,opt,name=op,proto3" json:"op,omitempty"`
	// Types that are valid to be assigned to Subject:
	//
	//	*Solve_File
	//	*Solve_Cache
	Subject       isSolve_Subject   `protobuf_oneof:"subject"`
	Description   map[string]string `protobuf:"bytes,6,rep,name=description,proto3" json:"description,omitempty" protobuf_key:"bytes,1,opt,name=key" protobuf_val:"bytes,2,opt,name=value"`
	unknownFields protoimpl.UnknownFields
	sizeCache     protoimpl.SizeCache
}

func (x *Solve) Reset() {
	*x = Solve{}
	mi := &file_github_com_moby_buildkit_solver_errdefs_errdefs_proto_msgTypes[6]
	ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
	ms.StoreMessageInfo(mi)
}

func (x *Solve) String() string {
	return protoimpl.X.MessageStringOf(x)
}

func (*Solve) ProtoMessage() {}

func (x *Solve) ProtoReflect() protoreflect.Message {
	mi := &file_github_com_moby_buildkit_solver_errdefs_errdefs_proto_msgTypes[6]
	if x != nil {
		ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
		if ms.LoadMessageInfo() == nil {
			ms.StoreMessageInfo(mi)
		}
		return ms
	}
	return mi.MessageOf(x)
}

// Deprecated: Use Solve.ProtoReflect.Descriptor instead.
func (*Solve) Descriptor() ([]byte, []int) {
	return file_github_com_moby_buildkit_solver_errdefs_errdefs_proto_rawDescGZIP(), []int{6}
}

func (x *Solve) GetInputIDs() []string {
	if x != nil {
		return x.InputIDs
	}
	return nil
}

func (x *Solve) GetMountIDs() []string {
	if x != nil {
		return x.MountIDs
	}
	return nil
}

func (x *Solve) GetOp() *pb.Op {
	if x != nil {
		return x.Op
	}
	return nil
}

func (x *Solve) GetSubject() isSolve_Subject {
	if x != nil {
		return x.Subject
	}
	return nil
}

func (x *Solve) GetFile() *FileAction {
	if x != nil {
		if x, ok := x.Subject.(*Solve_File); ok {
			return x.File
		}
	}
	return nil
}

func (x *Solve) GetCache() *ContentCache {
	if x != nil {
		if x, ok := x.Subject.(*Solve_Cache); ok {
			return x.Cache
		}
	}
	return nil
}

func (x *Solve) GetDescription() map[string]string {
	if x != nil {
		return x.Description
	}
	return nil
}

type isSolve_Subject interface {
	isSolve_Subject()
}

type Solve_File struct {
	File *FileAction `protobuf:"bytes,4,opt,name=file,proto3,oneof"`
}

type Solve_Cache struct {
	Cache *ContentCache `protobuf:"bytes,5,opt,name=cache,proto3,oneof"`
}

func (*Solve_File) isSolve_Subject() {}

func (*Solve_Cache) isSolve_Subject() {}

type FileAction struct {
	state protoimpl.MessageState `protogen:"open.v1"`
	// Index of the file action that failed the exec.
	Index         int64 `protobuf:"varint,1,opt,name=index,proto3" json:"index,omitempty"`
	unknownFields protoimpl.UnknownFields
	sizeCache     protoimpl.SizeCache
}

func (x *FileAction) Reset() {
	*x = FileAction{}
	mi := &file_github_com_moby_buildkit_solver_errdefs_errdefs_proto_msgTypes[7]
	ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
	ms.StoreMessageInfo(mi)
}

func (x *FileAction) String() string {
	return protoimpl.X.MessageStringOf(x)
}

func (*FileAction) ProtoMessage() {}

func (x *FileAction) ProtoReflect() protoreflect.Message {
	mi := &file_github_com_moby_buildkit_solver_errdefs_errdefs_proto_msgTypes[7]
	if x != nil {
		ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
		if ms.LoadMessageInfo() == nil {
			ms.StoreMessageInfo(mi)
		}
		return ms
	}
	return mi.MessageOf(x)
}

// Deprecated: Use FileAction.ProtoReflect.Descriptor instead.
func (*FileAction) Descriptor() ([]byte, []int) {
	return file_github_com_moby_buildkit_solver_errdefs_errdefs_proto_rawDescGZIP(), []int{7}
}

func (x *FileAction) GetIndex() int64 {
	if x != nil {
		return x.Index
	}
	return 0
}

type ContentCache struct {
	state protoimpl.MessageState `protogen:"open.v1"`
	// Original index of result that failed the slow cache calculation.
	Index         int64 `protobuf:"varint,1,opt,name=index,proto3" json:"index,omitempty"`
	unknownFields protoimpl.UnknownFields
	sizeCache     protoimpl.SizeCache
}

func (x *ContentCache) Reset() {
	*x = ContentCache{}
	mi := &file_github_com_moby_buildkit_solver_errdefs_errdefs_proto_msgTypes[8]
	ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
	ms.StoreMessageInfo(mi)
}

func (x *ContentCache) String() string {
	return protoimpl.X.MessageStringOf(x)
}

func (*ContentCache) ProtoMessage() {}

func (x *ContentCache) ProtoReflect() protoreflect.Message {
	mi := &file_github_com_moby_buildkit_solver_errdefs_errdefs_proto_msgTypes[8]
	if x != nil {
		ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
		if ms.LoadMessageInfo() == nil {
			ms.StoreMessageInfo(mi)
		}
		return ms
	}
	return mi.MessageOf(x)
}

// Deprecated: Use ContentCache.ProtoReflect.Descriptor instead.
func (*ContentCache) Descriptor() ([]byte, []int) {
	return file_github_com_moby_buildkit_solver_errdefs_errdefs_proto_rawDescGZIP(), []int{8}
}

func (x *ContentCache) GetIndex() int64 {
	if x != nil {
		return x.Index
	}
	return 0
}

type ProvenanceMaterialsIncomplete struct {
	state         protoimpl.MessageState          `protogen:"open.v1"`
	Incomplete    []*ProvenanceMaterialIncomplete `protobuf:"bytes,1,rep,name=incomplete,proto3" json:"incomplete,omitempty"`
	unknownFields protoimpl.UnknownFields
	sizeCache     protoimpl.SizeCache
}

func (x *ProvenanceMaterialsIncomplete) Reset() {
	*x = ProvenanceMaterialsIncomplete{}
	mi := &file_github_com_moby_buildkit_solver_errdefs_errdefs_proto_msgTypes[9]
	ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
	ms.StoreMessageInfo(mi)
}

func (x *ProvenanceMaterialsIncomplete) String() string {
	return protoimpl.X.MessageStringOf(x)
}

func (*ProvenanceMaterialsIncomplete) ProtoMessage() {}

func (x *ProvenanceMaterialsIncomplete) ProtoReflect() protoreflect.Message {
	mi := &file_github_com_moby_buildkit_solver_errdefs_errdefs_proto_msgTypes[9]
	if x != nil {
		ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
		if ms.LoadMessageInfo() == nil {
			ms.StoreMessageInfo(mi)
		}
		return ms
	}
	return mi.MessageOf(x)
}

// Deprecated: Use ProvenanceMaterialsIncomplete.ProtoReflect.Descriptor instead.
func (*ProvenanceMaterialsIncomplete) Descriptor() ([]byte, []int) {
	return file_github_com_moby_buildkit_solver_errdefs_errdefs_proto_rawDescGZIP(), []int{9}
}

func (x *ProvenanceMaterialsIncomplete) GetIncomplete() []*ProvenanceMaterialIncomplete {
	if x != nil {
		return x.Incomplete
	}
	return nil
}

type ProvenanceMaterialIncomplete struct {
	state         protoimpl.MessageState `protogen:"open.v1"`
	Op            string                 `protobuf:"bytes,1,opt,name=op,proto3" json:"op,omitempty"`
	Name          string                 `protobuf:"bytes,2,opt,name=name,proto3" json:"name,omitempty"`
	Method        string                 `protobuf:"bytes,3,opt,name=method,proto3" json:"method,omitempty"`
	Uri           string                 `protobuf:"bytes,4,opt,name=uri,proto3" json:"uri,omitempty"`
	FinalUri      string                 `protobuf:"bytes,5,opt,name=final_uri,json=finalUri,proto3" json:"final_uri,omitempty"`
	Reason        string                 `protobuf:"bytes,6,opt,name=reason,proto3" json:"reason,omitempty"`
	unknownFields protoimpl.UnknownFields
	sizeCache     protoimpl.SizeCache
}

func (x *ProvenanceMaterialIncomplete) Reset() {
	*x = ProvenanceMaterialIncomplete{}
	mi := &file_github_com_moby_buildkit_solver_errdefs_errdefs_proto_msgTypes[10]
	ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
	ms.StoreMessageInfo(mi)
}

func (x *ProvenanceMaterialIncomplete) String() string {
	return protoimpl.X.MessageStringOf(x)
}

func (*ProvenanceMaterialIncomplete) ProtoMessage() {}

func (x *ProvenanceMaterialIncomplete) ProtoReflect() protoreflect.Message {
	mi := &file_github_com_moby_buildkit_solver_errdefs_errdefs_proto_msgTypes[10]
	if x != nil {
		ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
		if ms.LoadMessageInfo() == nil {
			ms.StoreMessageInfo(mi)
		}
		return ms
	}
	return mi.MessageOf(x)
}

// Deprecated: Use ProvenanceMaterialIncomplete.ProtoReflect.Descriptor instead.
func (*ProvenanceMaterialIncomplete) Descriptor() ([]byte, []int) {
	return file_github_com_moby_buildkit_solver_errdefs_errdefs_proto_rawDescGZIP(), []int{10}
}

func (x *ProvenanceMaterialIncomplete) GetOp() string {
	if x != nil {
		return x.Op
	}
	return ""
}

func (x *ProvenanceMaterialIncomplete) GetName() string {
	if x != nil {
		return x.Name
	}
	return ""
}

func (x *ProvenanceMaterialIncomplete) GetMethod() string {
	if x != nil {
		return x.Method
	}
	return ""
}

func (x *ProvenanceMaterialIncomplete) GetUri() string {
	if x != nil {
		return x.Uri
	}
	return ""
}

func (x *ProvenanceMaterialIncomplete) GetFinalUri() string {
	if x != nil {
		return x.FinalUri
	}
	return ""
}

func (x *ProvenanceMaterialIncomplete) GetReason() string {
	if x != nil {
		return x.Reason
	}
	return ""
}

var File_github_com_moby_buildkit_solver_errdefs_errdefs_proto protoreflect.FileDescriptor

const file_github_com_moby_buildkit_solver_errdefs_errdefs_proto_rawDesc = "" +
	"\n" +
	"5github.com/moby/buildkit/solver/errdefs/errdefs.proto\x12\aerrdefs\x1a,github.com/moby/buildkit/solver/pb/ops.proto\" \n" +
	"\x06Vertex\x12\x16\n" +
	"\x06digest\x18\x01 \x01(\tR\x06digest\"O\n" +
	"\x06Source\x12\"\n" +
	"\x04info\x18\x01 \x01(\v2\x0e.pb.SourceInfoR\x04info\x12!\n" +
	"\x06ranges\x18\x02 \x03(\v2\t.pb.RangeR\x06ranges\"6\n" +
	"\bFrontend\x12\x12\n" +
	"\x04name\x18\x01 \x01(\tR\x04name\x12\x16\n" +
	"\x06source\x18\x02 \x01(\tR\x06source\"!\n" +
	"\vFrontendCap\x12\x12\n" +
	"\x04name\x18\x01 \x01(\tR\x04name\"J\n" +
	"\x14CompatibilityFeature\x12\x18\n" +
	"\aversion\x18\x01 \x01(\x03R\aversion\x12\x18\n" +
	"\afeature\x18\x02 \x01(\tR\afeature\" \n" +
	"\n" +
	"Subrequest\x12\x12\n" +
	"\x04name\x18\x01 \x01(\tR\x04name\"\xbf\x02\n" +
	"\x05Solve\x12\x1a\n" +
	"\binputIDs\x18\x01 \x03(\tR\binputIDs\x12\x1a\n" +
	"\bmountIDs\x18\x02 \x03(\tR\bmountIDs\x12\x16\n" +
	"\x02op\x18\x03 \x01(\v2\x06.pb.OpR\x02op\x12)\n" +
	"\x04file\x18\x04 \x01(\v2\x13.errdefs.FileActionH\x00R\x04file\x12-\n" +
	"\x05cache\x18\x05 \x01(\v2\x15.errdefs.ContentCacheH\x00R\x05cache\x12A\n" +
	"\vdescription\x18\x06 \x03(\v2\x1f.errdefs.Solve.DescriptionEntryR\vdescription\x1a>\n" +
	"\x10DescriptionEntry\x12\x10\n" +
	"\x03key\x18\x01 \x01(\tR\x03key\x12\x14\n" +
	"\x05value\x18\x02 \x01(\tR\x05value:\x028\x01B\t\n" +
	"\asubject\"\"\n" +
	"\n" +
	"FileAction\x12\x14\n" +
	"\x05index\x18\x01 \x01(\x03R\x05index\"$\n" +
	"\fContentCache\x12\x14\n" +
	"\x05index\x18\x01 \x01(\x03R\x05index\"f\n" +
	"\x1dProvenanceMaterialsIncomplete\x12E\n" +
	"\n" +
	"incomplete\x18\x01 \x03(\v2%.errdefs.ProvenanceMaterialIncompleteR\n" +
	"incomplete\"\xa1\x01\n" +
	"\x1cProvenanceMaterialIncomplete\x12\x0e\n" +
	"\x02op\x18\x01 \x01(\tR\x02op\x12\x12\n" +
	"\x04name\x18\x02 \x01(\tR\x04name\x12\x16\n" +
	"\x06method\x18\x03 \x01(\tR\x06method\x12\x10\n" +
	"\x03uri\x18\x04 \x01(\tR\x03uri\x12\x1b\n" +
	"\tfinal_uri\x18\x05 \x01(\tR\bfinalUri\x12\x16\n" +
	"\x06reason\x18\x06 \x01(\tR\x06reasonB)Z'github.com/moby/buildkit/solver/errdefsb\x06proto3"

var (
	file_github_com_moby_buildkit_solver_errdefs_errdefs_proto_rawDescOnce sync.Once
	file_github_com_moby_buildkit_solver_errdefs_errdefs_proto_rawDescData []byte
)

func file_github_com_moby_buildkit_solver_errdefs_errdefs_proto_rawDescGZIP() []byte {
	file_github_com_moby_buildkit_solver_errdefs_errdefs_proto_rawDescOnce.Do(func() {
		file_github_com_moby_buildkit_solver_errdefs_errdefs_proto_rawDescData = protoimpl.X.CompressGZIP(unsafe.Slice(unsafe.StringData(file_github_com_moby_buildkit_solver_errdefs_errdefs_proto_rawDesc), len(file_github_com_moby_buildkit_solver_errdefs_errdefs_proto_rawDesc)))
	})
	return file_github_com_moby_buildkit_solver_errdefs_errdefs_proto_rawDescData
}

var file_github_com_moby_buildkit_solver_errdefs_errdefs_proto_msgTypes = make([]protoimpl.MessageInfo, 12)
var file_github_com_moby_buildkit_solver_errdefs_errdefs_proto_goTypes = []any{
	(*Vertex)(nil),                        // 0: errdefs.Vertex
	(*Source)(nil),                        // 1: errdefs.Source
	(*Frontend)(nil),                      // 2: errdefs.Frontend
	(*FrontendCap)(nil),                   // 3: errdefs.FrontendCap
	(*CompatibilityFeature)(nil),          // 4: errdefs.CompatibilityFeature
	(*Subrequest)(nil),                    // 5: errdefs.Subrequest
	(*Solve)(nil),                         // 6: errdefs.Solve
	(*FileAction)(nil),                    // 7: errdefs.FileAction
	(*ContentCache)(nil),                  // 8: errdefs.ContentCache
	(*ProvenanceMaterialsIncomplete)(nil), // 9: errdefs.ProvenanceMaterialsIncomplete
	(*ProvenanceMaterialIncomplete)(nil),  // 10: errdefs.ProvenanceMaterialIncomplete
	nil,                                   // 11: errdefs.Solve.DescriptionEntry
	(*pb.SourceInfo)(nil),                 // 12: pb.SourceInfo
	(*pb.Range)(nil),                      // 13: pb.Range
	(*pb.Op)(nil),                         // 14: pb.Op
}
var file_github_com_moby_buildkit_solver_errdefs_errdefs_proto_depIdxs = []int32{
	12, // 0: errdefs.Source.info:type_name -> pb.SourceInfo
	13, // 1: errdefs.Source.ranges:type_name -> pb.Range
	14, // 2: errdefs.Solve.op:type_name -> pb.Op
	7,  // 3: errdefs.Solve.file:type_name -> errdefs.FileAction
	8,  // 4: errdefs.Solve.cache:type_name -> errdefs.ContentCache
	11, // 5: errdefs.Solve.description:type_name -> errdefs.Solve.DescriptionEntry
	10, // 6: errdefs.ProvenanceMaterialsIncomplete.incomplete:type_name -> errdefs.ProvenanceMaterialIncomplete
	7,  // [7:7] is the sub-list for method output_type
	7,  // [7:7] is the sub-list for method input_type
	7,  // [7:7] is the sub-list for extension type_name
	7,  // [7:7] is the sub-list for extension extendee
	0,  // [0:7] is the sub-list for field type_name
}

func init() { file_github_com_moby_buildkit_solver_errdefs_errdefs_proto_init() }
func file_github_com_moby_buildkit_solver_errdefs_errdefs_proto_init() {
	if File_github_com_moby_buildkit_solver_errdefs_errdefs_proto != nil {
		return
	}
	file_github_com_moby_buildkit_solver_errdefs_errdefs_proto_msgTypes[6].OneofWrappers = []any{
		(*Solve_File)(nil),
		(*Solve_Cache)(nil),
	}
	type x struct{}
	out := protoimpl.TypeBuilder{
		File: protoimpl.DescBuilder{
			GoPackagePath: reflect.TypeOf(x{}).PkgPath(),
			RawDescriptor: unsafe.Slice(unsafe.StringData(file_github_com_moby_buildkit_solver_errdefs_errdefs_proto_rawDesc), len(file_github_com_moby_buildkit_solver_errdefs_errdefs_proto_rawDesc)),
			NumEnums:      0,
			NumMessages:   12,
			NumExtensions: 0,
			NumServices:   0,
		},
		GoTypes:           file_github_com_moby_buildkit_solver_errdefs_errdefs_proto_goTypes,
		DependencyIndexes: file_github_com_moby_buildkit_solver_errdefs_errdefs_proto_depIdxs,
		MessageInfos:      file_github_com_moby_buildkit_solver_errdefs_errdefs_proto_msgTypes,
	}.Build()
	File_github_com_moby_buildkit_solver_errdefs_errdefs_proto = out.File
	file_github_com_moby_buildkit_solver_errdefs_errdefs_proto_goTypes = nil
	file_github_com_moby_buildkit_solver_errdefs_errdefs_proto_depIdxs = nil
}
