// Code generated by protoc-gen-go. DO NOT EDIT.
// versions:
// 	protoc-gen-go v1.36.11
// 	protoc        v3.14.0
// source: github.com/moby/buildkit/session/exporter/exporter.proto

package exporter

import (
	protoreflect "google.golang.org/protobuf/reflect/protoreflect"
	protoimpl "google.golang.org/protobuf/runtime/protoimpl"
	reflect "reflect"
	sync "sync"
	unsafe "unsafe"
)

const (
	// Verify that this generated code is sufficiently up-to-date.
	_ = protoimpl.EnforceVersion(20 - protoimpl.MinVersion)
	// Verify that runtime/protoimpl is sufficiently up-to-date.
	_ = protoimpl.EnforceVersion(protoimpl.MaxVersion - 20)
)

type FindExportersRequest struct {
	state         protoimpl.MessageState `protogen:"open.v1"`
	Metadata      map[string][]byte      `protobuf:"bytes,1,rep,name=metadata,proto3" json:"metadata,omitempty" protobuf_key:"bytes,1,opt,name=key" protobuf_val:"bytes,2,opt,name=value"`
	Refs          []string               `protobuf:"bytes,2,rep,name=refs,proto3" json:"refs,omitempty"`
	unknownFields protoimpl.UnknownFields
	sizeCache     protoimpl.SizeCache
}

func (x *FindExportersRequest) Reset() {
	*x = FindExportersRequest{}
	mi := &file_github_com_moby_buildkit_session_exporter_exporter_proto_msgTypes[0]
	ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
	ms.StoreMessageInfo(mi)
}

func (x *FindExportersRequest) String() string {
	return protoimpl.X.MessageStringOf(x)
}

func (*FindExportersRequest) ProtoMessage() {}

func (x *FindExportersRequest) ProtoReflect() protoreflect.Message {
	mi := &file_github_com_moby_buildkit_session_exporter_exporter_proto_msgTypes[0]
	if x != nil {
		ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
		if ms.LoadMessageInfo() == nil {
			ms.StoreMessageInfo(mi)
		}
		return ms
	}
	return mi.MessageOf(x)
}

// Deprecated: Use FindExportersRequest.ProtoReflect.Descriptor instead.
func (*FindExportersRequest) Descriptor() ([]byte, []int) {
	return file_github_com_moby_buildkit_session_exporter_exporter_proto_rawDescGZIP(), []int{0}
}

func (x *FindExportersRequest) GetMetadata() map[string][]byte {
	if x != nil {
		return x.Metadata
	}
	return nil
}

func (x *FindExportersRequest) GetRefs() []string {
	if x != nil {
		return x.Refs
	}
	return nil
}

type FindExportersResponse struct {
	state         protoimpl.MessageState `protogen:"open.v1"`
	Exporters     []*ExporterRequest     `protobuf:"bytes,1,rep,name=exporters,proto3" json:"exporters,omitempty"`
	unknownFields protoimpl.UnknownFields
	sizeCache     protoimpl.SizeCache
}

func (x *FindExportersResponse) Reset() {
	*x = FindExportersResponse{}
	mi := &file_github_com_moby_buildkit_session_exporter_exporter_proto_msgTypes[1]
	ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
	ms.StoreMessageInfo(mi)
}

func (x *FindExportersResponse) String() string {
	return protoimpl.X.MessageStringOf(x)
}

func (*FindExportersResponse) ProtoMessage() {}

func (x *FindExportersResponse) ProtoReflect() protoreflect.Message {
	mi := &file_github_com_moby_buildkit_session_exporter_exporter_proto_msgTypes[1]
	if x != nil {
		ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
		if ms.LoadMessageInfo() == nil {
			ms.StoreMessageInfo(mi)
		}
		return ms
	}
	return mi.MessageOf(x)
}

// Deprecated: Use FindExportersResponse.ProtoReflect.Descriptor instead.
func (*FindExportersResponse) Descriptor() ([]byte, []int) {
	return file_github_com_moby_buildkit_session_exporter_exporter_proto_rawDescGZIP(), []int{1}
}

func (x *FindExportersResponse) GetExporters() []*ExporterRequest {
	if x != nil {
		return x.Exporters
	}
	return nil
}

type ExporterRequest struct {
	state         protoimpl.MessageState `protogen:"open.v1"`
	Type          string                 `protobuf:"bytes,1,opt,name=Type,proto3" json:"Type,omitempty"`
	Attrs         map[string]string      `protobuf:"bytes,2,rep,name=Attrs,proto3" json:"Attrs,omitempty" protobuf_key:"bytes,1,opt,name=key" protobuf_val:"bytes,2,opt,name=value"`
	unknownFields protoimpl.UnknownFields
	sizeCache     protoimpl.SizeCache
}

func (x *ExporterRequest) Reset() {
	*x = ExporterRequest{}
	mi := &file_github_com_moby_buildkit_session_exporter_exporter_proto_msgTypes[2]
	ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
	ms.StoreMessageInfo(mi)
}

func (x *ExporterRequest) String() string {
	return protoimpl.X.MessageStringOf(x)
}

func (*ExporterRequest) ProtoMessage() {}

func (x *ExporterRequest) ProtoReflect() protoreflect.Message {
	mi := &file_github_com_moby_buildkit_session_exporter_exporter_proto_msgTypes[2]
	if x != nil {
		ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
		if ms.LoadMessageInfo() == nil {
			ms.StoreMessageInfo(mi)
		}
		return ms
	}
	return mi.MessageOf(x)
}

// Deprecated: Use ExporterRequest.ProtoReflect.Descriptor instead.
func (*ExporterRequest) Descriptor() ([]byte, []int) {
	return file_github_com_moby_buildkit_session_exporter_exporter_proto_rawDescGZIP(), []int{2}
}

func (x *ExporterRequest) GetType() string {
	if x != nil {
		return x.Type
	}
	return ""
}

func (x *ExporterRequest) GetAttrs() map[string]string {
	if x != nil {
		return x.Attrs
	}
	return nil
}

var File_github_com_moby_buildkit_session_exporter_exporter_proto protoreflect.FileDescriptor

const file_github_com_moby_buildkit_session_exporter_exporter_proto_rawDesc = "" +
	"\n" +
	"8github.com/moby/buildkit/session/exporter/exporter.proto\x12\x10moby.exporter.v1\"\xb9\x01\n" +
	"\x14FindExportersRequest\x12P\n" +
	"\bmetadata\x18\x01 \x03(\v24.moby.exporter.v1.FindExportersRequest.MetadataEntryR\bmetadata\x12\x12\n" +
	"\x04refs\x18\x02 \x03(\tR\x04refs\x1a;\n" +
	"\rMetadataEntry\x12\x10\n" +
	"\x03key\x18\x01 \x01(\tR\x03key\x12\x14\n" +
	"\x05value\x18\x02 \x01(\fR\x05value:\x028\x01\"X\n" +
	"\x15FindExportersResponse\x12?\n" +
	"\texporters\x18\x01 \x03(\v2!.moby.exporter.v1.ExporterRequestR\texporters\"\xa3\x01\n" +
	"\x0fExporterRequest\x12\x12\n" +
	"\x04Type\x18\x01 \x01(\tR\x04Type\x12B\n" +
	"\x05Attrs\x18\x02 \x03(\v2,.moby.exporter.v1.ExporterRequest.AttrsEntryR\x05Attrs\x1a8\n" +
	"\n" +
	"AttrsEntry\x12\x10\n" +
	"\x03key\x18\x01 \x01(\tR\x03key\x12\x14\n" +
	"\x05value\x18\x02 \x01(\tR\x05value:\x028\x012l\n" +
	"\bExporter\x12`\n" +
	"\rFindExporters\x12&.moby.exporter.v1.FindExportersRequest\x1a'.moby.exporter.v1.FindExportersResponseB+Z)github.com/moby/buildkit/session/exporterb\x06proto3"

var (
	file_github_com_moby_buildkit_session_exporter_exporter_proto_rawDescOnce sync.Once
	file_github_com_moby_buildkit_session_exporter_exporter_proto_rawDescData []byte
)

func file_github_com_moby_buildkit_session_exporter_exporter_proto_rawDescGZIP() []byte {
	file_github_com_moby_buildkit_session_exporter_exporter_proto_rawDescOnce.Do(func() {
		file_github_com_moby_buildkit_session_exporter_exporter_proto_rawDescData = protoimpl.X.CompressGZIP(unsafe.Slice(unsafe.StringData(file_github_com_moby_buildkit_session_exporter_exporter_proto_rawDesc), len(file_github_com_moby_buildkit_session_exporter_exporter_proto_rawDesc)))
	})
	return file_github_com_moby_buildkit_session_exporter_exporter_proto_rawDescData
}

var file_github_com_moby_buildkit_session_exporter_exporter_proto_msgTypes = make([]protoimpl.MessageInfo, 5)
var file_github_com_moby_buildkit_session_exporter_exporter_proto_goTypes = []any{
	(*FindExportersRequest)(nil),  // 0: moby.exporter.v1.FindExportersRequest
	(*FindExportersResponse)(nil), // 1: moby.exporter.v1.FindExportersResponse
	(*ExporterRequest)(nil),       // 2: moby.exporter.v1.ExporterRequest
	nil,                           // 3: moby.exporter.v1.FindExportersRequest.MetadataEntry
	nil,                           // 4: moby.exporter.v1.ExporterRequest.AttrsEntry
}
var file_github_com_moby_buildkit_session_exporter_exporter_proto_depIdxs = []int32{
	3, // 0: moby.exporter.v1.FindExportersRequest.metadata:type_name -> moby.exporter.v1.FindExportersRequest.MetadataEntry
	2, // 1: moby.exporter.v1.FindExportersResponse.exporters:type_name -> moby.exporter.v1.ExporterRequest
	4, // 2: moby.exporter.v1.ExporterRequest.Attrs:type_name -> moby.exporter.v1.ExporterRequest.AttrsEntry
	0, // 3: moby.exporter.v1.Exporter.FindExporters:input_type -> moby.exporter.v1.FindExportersRequest
	1, // 4: moby.exporter.v1.Exporter.FindExporters:output_type -> moby.exporter.v1.FindExportersResponse
	4, // [4:5] is the sub-list for method output_type
	3, // [3:4] is the sub-list for method input_type
	3, // [3:3] is the sub-list for extension type_name
	3, // [3:3] is the sub-list for extension extendee
	0, // [0:3] is the sub-list for field type_name
}

func init() { file_github_com_moby_buildkit_session_exporter_exporter_proto_init() }
func file_github_com_moby_buildkit_session_exporter_exporter_proto_init() {
	if File_github_com_moby_buildkit_session_exporter_exporter_proto != nil {
		return
	}
	type x struct{}
	out := protoimpl.TypeBuilder{
		File: protoimpl.DescBuilder{
			GoPackagePath: reflect.TypeOf(x{}).PkgPath(),
			RawDescriptor: unsafe.Slice(unsafe.StringData(file_github_com_moby_buildkit_session_exporter_exporter_proto_rawDesc), len(file_github_com_moby_buildkit_session_exporter_exporter_proto_rawDesc)),
			NumEnums:      0,
			NumMessages:   5,
			NumExtensions: 0,
			NumServices:   1,
		},
		GoTypes:           file_github_com_moby_buildkit_session_exporter_exporter_proto_goTypes,
		DependencyIndexes: file_github_com_moby_buildkit_session_exporter_exporter_proto_depIdxs,
		MessageInfos:      file_github_com_moby_buildkit_session_exporter_exporter_proto_msgTypes,
	}.Build()
	File_github_com_moby_buildkit_session_exporter_exporter_proto = out.File
	file_github_com_moby_buildkit_session_exporter_exporter_proto_goTypes = nil
	file_github_com_moby_buildkit_session_exporter_exporter_proto_depIdxs = nil
}
