package nl

const (
	VDPA_GENL_NAME    = "vdpa"
	VDPA_GENL_VERSION = 0x1
)

const (
	VDPA_CMD_UNSPEC = iota
	VDPA_CMD_MGMTDEV_NEW
	VDPA_CMD_MGMTDEV_GET /* can dump */
	VDPA_CMD_DEV_NEW
	VDPA_CMD_DEV_DEL
	VDPA_CMD_DEV_GET        /* can dump */
	VDPA_CMD_DEV_CONFIG_GET /* can dump */
	VDPA_CMD_DEV_VSTATS_GET
)

const (
	VDPA_ATTR_UNSPEC = iota
	VDPA_ATTR_MGMTDEV_BUS_NAME
	VDPA_ATTR_MGMTDEV_DEV_NAME
	VDPA_ATTR_MGMTDEV_SUPPORTED_CLASSES
	VDPA_ATTR_DEV_NAME
	VDPA_ATTR_DEV_ID
	VDPA_ATTR_DEV_VENDOR_ID
	VDPA_ATTR_DEV_MAX_VQS
	VDPA_ATTR_DEV_MAX_VQ_SIZE
	VDPA_ATTR_DEV_MIN_VQ_SIZE
	VDPA_ATTR_DEV_NET_CFG_MACADDR
	VDPA_ATTR_DEV_NET_STATUS
	VDPA_ATTR_DEV_NET_CFG_MAX_VQP
	VDPA_ATTR_DEV_NET_CFG_MTU
	VDPA_ATTR_DEV_NEGOTIATED_FEATURES
	VDPA_ATTR_DEV_MGMTDEV_MAX_VQS
	VDPA_ATTR_DEV_SUPPORTED_FEATURES
	VDPA_ATTR_DEV_QUEUE_INDEX
	VDPA_ATTR_DEV_VENDOR_ATTR_NAME
	VDPA_ATTR_DEV_VENDOR_ATTR_VALUE
	VDPA_ATTR_DEV_FEATURES
)
