// Code generated by protoc-gen-go. DO NOT EDIT.
// versions:
// 	protoc-gen-go v1.36.11
// 	protoc        v3.11.4
// source: github.com/tonistiigi/fsutil/types/wire.proto

package types

import (
	_ "github.com/planetscale/vtprotobuf/vtproto"
	protoreflect "google.golang.org/protobuf/reflect/protoreflect"
	protoimpl "google.golang.org/protobuf/runtime/protoimpl"
	reflect "reflect"
	sync "sync"
	unsafe "unsafe"
)

const (
	// Verify that this generated code is sufficiently up-to-date.
	_ = protoimpl.EnforceVersion(20 - protoimpl.MinVersion)
	// Verify that runtime/protoimpl is sufficiently up-to-date.
	_ = protoimpl.EnforceVersion(protoimpl.MaxVersion - 20)
)

type Packet_PacketType int32

const (
	Packet_PACKET_STAT Packet_PacketType = 0
	Packet_PACKET_REQ  Packet_PacketType = 1
	Packet_PACKET_DATA Packet_PacketType = 2
	Packet_PACKET_FIN  Packet_PacketType = 3
	Packet_PACKET_ERR  Packet_PacketType = 4
)

// Enum value maps for Packet_PacketType.
var (
	Packet_PacketType_name = map[int32]string{
		0: "PACKET_STAT",
		1: "PACKET_REQ",
		2: "PACKET_DATA",
		3: "PACKET_FIN",
		4: "PACKET_ERR",
	}
	Packet_PacketType_value = map[string]int32{
		"PACKET_STAT": 0,
		"PACKET_REQ":  1,
		"PACKET_DATA": 2,
		"PACKET_FIN":  3,
		"PACKET_ERR":  4,
	}
)

func (x Packet_PacketType) Enum() *Packet_PacketType {
	p := new(Packet_PacketType)
	*p = x
	return p
}

func (x Packet_PacketType) String() string {
	return protoimpl.X.EnumStringOf(x.Descriptor(), protoreflect.EnumNumber(x))
}

func (Packet_PacketType) Descriptor() protoreflect.EnumDescriptor {
	return file_github_com_tonistiigi_fsutil_types_wire_proto_enumTypes[0].Descriptor()
}

func (Packet_PacketType) Type() protoreflect.EnumType {
	return &file_github_com_tonistiigi_fsutil_types_wire_proto_enumTypes[0]
}

func (x Packet_PacketType) Number() protoreflect.EnumNumber {
	return protoreflect.EnumNumber(x)
}

// Deprecated: Use Packet_PacketType.Descriptor instead.
func (Packet_PacketType) EnumDescriptor() ([]byte, []int) {
	return file_github_com_tonistiigi_fsutil_types_wire_proto_rawDescGZIP(), []int{0, 0}
}

type Packet struct {
	state         protoimpl.MessageState `protogen:"open.v1"`
	Type          Packet_PacketType      `protobuf:"varint,1,opt,name=type,proto3,enum=fsutil.types.Packet_PacketType" json:"type,omitempty"`
	Stat          *Stat                  `protobuf:"bytes,2,opt,name=stat,proto3" json:"stat,omitempty"`
	ID            uint32                 `protobuf:"varint,3,opt,name=ID,proto3" json:"ID,omitempty"`
	Data          []byte                 `protobuf:"bytes,4,opt,name=data,proto3" json:"data,omitempty"`
	unknownFields protoimpl.UnknownFields
	sizeCache     protoimpl.SizeCache
}

func (x *Packet) Reset() {
	*x = Packet{}
	mi := &file_github_com_tonistiigi_fsutil_types_wire_proto_msgTypes[0]
	ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
	ms.StoreMessageInfo(mi)
}

func (x *Packet) String() string {
	return protoimpl.X.MessageStringOf(x)
}

func (*Packet) ProtoMessage() {}

func (x *Packet) ProtoReflect() protoreflect.Message {
	mi := &file_github_com_tonistiigi_fsutil_types_wire_proto_msgTypes[0]
	if x != nil {
		ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
		if ms.LoadMessageInfo() == nil {
			ms.StoreMessageInfo(mi)
		}
		return ms
	}
	return mi.MessageOf(x)
}

// Deprecated: Use Packet.ProtoReflect.Descriptor instead.
func (*Packet) Descriptor() ([]byte, []int) {
	return file_github_com_tonistiigi_fsutil_types_wire_proto_rawDescGZIP(), []int{0}
}

func (x *Packet) GetType() Packet_PacketType {
	if x != nil {
		return x.Type
	}
	return Packet_PACKET_STAT
}

func (x *Packet) GetStat() *Stat {
	if x != nil {
		return x.Stat
	}
	return nil
}

func (x *Packet) GetID() uint32 {
	if x != nil {
		return x.ID
	}
	return 0
}

func (x *Packet) GetData() []byte {
	if x != nil {
		return x.Data
	}
	return nil
}

var File_github_com_tonistiigi_fsutil_types_wire_proto protoreflect.FileDescriptor

const file_github_com_tonistiigi_fsutil_types_wire_proto_rawDesc = "" +
	"\n" +
	"-github.com/tonistiigi/fsutil/types/wire.proto\x12\ffsutil.types\x1a3github.com/planetscale/vtprotobuf/vtproto/ext.proto\x1a-github.com/tonistiigi/fsutil/types/stat.proto\"\xef\x01\n" +
	"\x06Packet\x123\n" +
	"\x04type\x18\x01 \x01(\x0e2\x1f.fsutil.types.Packet.PacketTypeR\x04type\x12&\n" +
	"\x04stat\x18\x02 \x01(\v2\x12.fsutil.types.StatR\x04stat\x12\x0e\n" +
	"\x02ID\x18\x03 \x01(\rR\x02ID\x12\x12\n" +
	"\x04data\x18\x04 \x01(\fR\x04data\"^\n" +
	"\n" +
	"PacketType\x12\x0f\n" +
	"\vPACKET_STAT\x10\x00\x12\x0e\n" +
	"\n" +
	"PACKET_REQ\x10\x01\x12\x0f\n" +
	"\vPACKET_DATA\x10\x02\x12\x0e\n" +
	"\n" +
	"PACKET_FIN\x10\x03\x12\x0e\n" +
	"\n" +
	"PACKET_ERR\x10\x04:\x04\xa8\xa6\x1f\x01B$Z\"github.com/tonistiigi/fsutil/typesb\x06proto3"

var (
	file_github_com_tonistiigi_fsutil_types_wire_proto_rawDescOnce sync.Once
	file_github_com_tonistiigi_fsutil_types_wire_proto_rawDescData []byte
)

func file_github_com_tonistiigi_fsutil_types_wire_proto_rawDescGZIP() []byte {
	file_github_com_tonistiigi_fsutil_types_wire_proto_rawDescOnce.Do(func() {
		file_github_com_tonistiigi_fsutil_types_wire_proto_rawDescData = protoimpl.X.CompressGZIP(unsafe.Slice(unsafe.StringData(file_github_com_tonistiigi_fsutil_types_wire_proto_rawDesc), len(file_github_com_tonistiigi_fsutil_types_wire_proto_rawDesc)))
	})
	return file_github_com_tonistiigi_fsutil_types_wire_proto_rawDescData
}

var file_github_com_tonistiigi_fsutil_types_wire_proto_enumTypes = make([]protoimpl.EnumInfo, 1)
var file_github_com_tonistiigi_fsutil_types_wire_proto_msgTypes = make([]protoimpl.MessageInfo, 1)
var file_github_com_tonistiigi_fsutil_types_wire_proto_goTypes = []any{
	(Packet_PacketType)(0), // 0: fsutil.types.Packet.PacketType
	(*Packet)(nil),         // 1: fsutil.types.Packet
	(*Stat)(nil),           // 2: fsutil.types.Stat
}
var file_github_com_tonistiigi_fsutil_types_wire_proto_depIdxs = []int32{
	0, // 0: fsutil.types.Packet.type:type_name -> fsutil.types.Packet.PacketType
	2, // 1: fsutil.types.Packet.stat:type_name -> fsutil.types.Stat
	2, // [2:2] is the sub-list for method output_type
	2, // [2:2] is the sub-list for method input_type
	2, // [2:2] is the sub-list for extension type_name
	2, // [2:2] is the sub-list for extension extendee
	0, // [0:2] is the sub-list for field type_name
}

func init() { file_github_com_tonistiigi_fsutil_types_wire_proto_init() }
func file_github_com_tonistiigi_fsutil_types_wire_proto_init() {
	if File_github_com_tonistiigi_fsutil_types_wire_proto != nil {
		return
	}
	file_github_com_tonistiigi_fsutil_types_stat_proto_init()
	type x struct{}
	out := protoimpl.TypeBuilder{
		File: protoimpl.DescBuilder{
			GoPackagePath: reflect.TypeOf(x{}).PkgPath(),
			RawDescriptor: unsafe.Slice(unsafe.StringData(file_github_com_tonistiigi_fsutil_types_wire_proto_rawDesc), len(file_github_com_tonistiigi_fsutil_types_wire_proto_rawDesc)),
			NumEnums:      1,
			NumMessages:   1,
			NumExtensions: 0,
			NumServices:   0,
		},
		GoTypes:           file_github_com_tonistiigi_fsutil_types_wire_proto_goTypes,
		DependencyIndexes: file_github_com_tonistiigi_fsutil_types_wire_proto_depIdxs,
		EnumInfos:         file_github_com_tonistiigi_fsutil_types_wire_proto_enumTypes,
		MessageInfos:      file_github_com_tonistiigi_fsutil_types_wire_proto_msgTypes,
	}.Build()
	File_github_com_tonistiigi_fsutil_types_wire_proto = out.File
	file_github_com_tonistiigi_fsutil_types_wire_proto_goTypes = nil
	file_github_com_tonistiigi_fsutil_types_wire_proto_depIdxs = nil
}
