// Copyright The OpenTelemetry Authors
// SPDX-License-Identifier: Apache-2.0

package internal // import "go.opentelemetry.io/otel/exporters/otlp/otlptrace/otlptracegrpc/internal"

// Version is the current release version of the OpenTelemetry OTLP gRPC trace
// exporter in use.
const Version = "1.44.0"
